from wandb.plots.explain_text import explain_text as ExplainText
from wandb.plots.heatmap import heatmap as HeatMap
from wandb.plots.named_entity import named_entity as NER
from wandb.plots.part_of_speech import part_of_speech as POS
from wandb.plots.precision_recall import precision_recall
from wandb.plots.roc import roc as ROC
